(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2013-2015)
Author:  Julian Nagele <julian.nagele@uibk.ac.at> (2013-2016)
Author:  Makarius Wenzel <makarius@sketis.net> (2013)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2014, 2015)
License: LGPL (see file COPYING.LESSER)
*)
theory Strongly_Closed
imports
  "$AFP/Abstract-Rewriting/Abstract_Rewriting"
  Critical_Pairs
begin

definition strongly_closed :: "('f,string) trs \<Rightarrow> bool"
where
  "strongly_closed R = (\<forall> (b, p, q) \<in> critical_pairs R R.
    \<exists>s r. (p, r) \<in> (rstep R)\<^sup>* \<and> (q, r) \<in> (rstep R)\<^sup>=
        \<and> (p, s) \<in> (rstep R)\<^sup>= \<and> (q, s) \<in> (rstep R)\<^sup>*)"

lemma strongly_closed_nested_pos:
  fixes l1 :: "('a, char list) Term.term"
  assumes "(s, t) \<in> rstep_r_p_s R (l1, r1) p1 \<sigma>1"
    and "(s, u) \<in> rstep_r_p_s R (l2, r2) p2 \<sigma>2"
    and "p1 \<le> p2"
    and sc:"strongly_closed R"
    and lin: "linear_trs R"
  shows "\<exists>v w. (t, v) \<in> (rstep R)\<^sup>* \<and> (u, v) \<in> (rstep R)\<^sup>= \<and> (t, w) \<in> (rstep R)\<^sup>= \<and> (u, w) \<in> (rstep R)\<^sup>*"
proof -
  let ?R = "rstep R"
  from assms have p1: "p1 \<in> poss s" and \<sigma>1: "s |_ p1 = l1 \<cdot> \<sigma>1"
  and t: "t = replace_at s p1 (r1 \<cdot> \<sigma>1)" and lr1: "(l1, r1) \<in> R" and p2: "p2 \<in> poss s"
  and \<sigma>2: "s |_ p2 = l2 \<cdot> \<sigma>2" and u: "u = replace_at s p2 (r2 \<cdot> \<sigma>2)" and lr2: "(l2, r2) \<in> R"
    unfolding rstep_r_p_s_def' by auto
  from assms obtain q where q:"p2 = p1 <#> q" using less_eq_pos_def by auto
  then have l2q:"(l1 \<cdot> \<sigma>1) |_ q = l2 \<cdot> \<sigma>2" using \<sigma>1 \<sigma>2 p1 p2 by auto
  from q u \<sigma>1 have up1:"u = replace_at s p1 (replace_at (l1 \<cdot> \<sigma>1) q (r2 \<cdot> \<sigma>2))"
    using ctxt_of_pos_term_append[OF p1] ctxt_compose_apply_distrib by auto
  show "\<exists>v w. (t, v) \<in> (rstep R)\<^sup>* \<and> (u, v) \<in> (rstep R)\<^sup>= \<and> (t, w) \<in> (rstep R)\<^sup>= \<and> (u, w) \<in> (rstep R)\<^sup>*"
  proof (cases "q \<in> funposs l1")
    case True
    then have ql1:"q \<in> poss l1" and l1q:"is_Fun (l1 |_ q)" and l1qs1: "l1 |_ q \<cdot> \<sigma>1 = l2 \<cdot> \<sigma>2"
      using funposs_fun_conv[OF True] funposs_imp_poss[OF True] l2q subt_at_subst by auto
    from mgu_var_disjoint_string_complete[OF l1qs1]
    obtain \<mu>1 \<mu>2 \<delta> where mgu: "mgu_var_disjoint_string (l1 |_ q) l2 = Some (\<mu>1,\<mu>2)"
      and d1:"\<sigma>1 = \<mu>1 \<circ>\<^sub>s \<delta>"  and d2:"\<sigma>2 = \<mu>2 \<circ>\<^sub>s \<delta>" and mu12:"l1 |_ q \<cdot> \<mu>1 = l2 \<cdot> \<mu>2"
      by auto
    def t' \<equiv> "r1 \<cdot> \<mu>1"
    def u' \<equiv> "(ctxt_of_pos_term q l1 \<cdot>\<^sub>c \<mu>1)\<langle>r2 \<cdot> \<mu>2\<rangle>"
    from critical_pairsI[OF lr1 lr2 _ l1q mgu] ctxt_supt_id[OF ql1]
    have "(ctxt_of_pos_term q l1 = \<box>, t', u') \<in> critical_pairs R R"
      unfolding t'_def u'_def by auto
    with sc[unfolded strongly_closed_def] obtain v w where
      v: "(t', v) \<in> ?R\<^sup>* \<and> (u', v) \<in> ?R\<^sup>=" and n: "(t', w) \<in> ?R\<^sup>= \<and> (u', w) \<in> ?R\<^sup>*"
      by blast
    from d2 d1 have d2:"u' \<cdot> \<delta> = replace_at (l1 \<cdot> \<sigma>1) q (r2 \<cdot> \<sigma>2)"
      unfolding u'_def using ctxt_compose_subst_compose_distrib ctxt_of_pos_term_subst[OF ql1]
        subst_apply_term_ctxt_apply_distrib subst_compose_apply_term_distrib by metis
    from d1 t'_def have "t' \<cdot> \<delta> = r1 \<cdot> \<sigma>1" by auto
    with t have tt':"t = replace_at s p1 (t' \<cdot> \<delta>)" by auto
    from d2 u up1 have uu':"u = replace_at s p1 (u' \<cdot> \<delta>)" by auto
    from v have msteps:"(t' \<cdot> \<delta>, v \<cdot> \<delta>) \<in> ?R\<^sup>* \<and> (u' \<cdot> \<delta>, v \<cdot> \<delta>) \<in> ?R\<^sup>="
    using rstep.subst rstep_id rstep_union rsteps_closed_subst by metis
    from n have nsteps:"(t' \<cdot> \<delta>, w \<cdot> \<delta>) \<in> ?R\<^sup>= \<and> (u' \<cdot> \<delta>, w \<cdot> \<delta>) \<in> ?R\<^sup>*"
      using rstep.subst rstep_id rstep_union rsteps_closed_subst by metis
    from msteps tt' have "(t, replace_at s p1 (v \<cdot> \<delta>)) \<in> ?R\<^sup>*" using rsteps_closed_ctxt by auto
    moreover from msteps uu' have "(u, replace_at s p1 (v \<cdot> \<delta>)) \<in> ?R\<^sup>=" by auto
    moreover from nsteps tt' have "(t, replace_at s p1 (w \<cdot> \<delta>)) \<in> ?R\<^sup>=" by auto
    moreover from nsteps uu' have "(u, replace_at s p1 (w \<cdot> \<delta>)) \<in> ?R\<^sup>*" using rsteps_closed_ctxt by auto
    ultimately show ?thesis by blast
  next
    case False
    then have "\<not>(q \<in> poss l1 \<and> is_Fun (l1 |_ q))"
      by (metis DiffI is_Var_def poss_simps(3) varposs_iff)
    then obtain q1 q2 x where  q12:"q = q1 <#> q2" and q1:"q1 \<in> poss l1" and x:"l1 |_ q1 = Var x"
      using \<sigma>1 p2 pos_into_subst[of l1 \<sigma>1 "s |_ p1" q] poss_append_poss q by auto
    then have xq2:"\<sigma>1 x |_ q2 = l2 \<cdot> \<sigma>2"
      using l2q subt_at_append[OF poss_imp_subst_poss[OF q1, of \<sigma>1]] by auto
    have q1l1\<sigma>1:"q1 \<in> poss (l1 \<cdot> \<sigma>1)" using q1 by simp
    def \<tau> \<equiv> "\<lambda> y. if y = x then replace_at (\<sigma>1 x) q2 (r2 \<cdot> \<sigma>2) else \<sigma>1 y"
    let ?cr = "replace_at s p1 (r1 \<cdot> \<tau>)"
    from lin lr1 have linl1:"linear_term l1" and linr1: "linear_term r1" by (auto dest: linear_trsE)
    from linear_term_replace_in_subst[OF linl1 q1 x, of \<sigma>1 \<tau>]
    have "l1 \<cdot> \<tau> = replace_at (l1 \<cdot> \<sigma>1) q1 (\<tau> x)" by (simp add: \<tau>_def)
    also have "... = replace_at (l1 \<cdot> \<sigma>1) q (r2 \<cdot> \<sigma>2)"
      using q12 q1 x ctxt_of_pos_term_append[OF q1l1\<sigma>1] ctxt_compose_apply_distrib \<tau>_def by auto
    finally have "l1 \<cdot> \<tau> = replace_at (l1 \<cdot> \<sigma>1) q (r2 \<cdot> \<sigma>2)" .
    with up1 have "u = replace_at s p1 (l1 \<cdot> \<tau>)" by auto
    then have ucr:"(u, ?cr) \<in> ?R" using subset_rstep lr1 by auto
    have tcr:"(t, ?cr) \<in> ?R\<^sup>="
    proof (cases "x \<in> vars_term r1")
      case False
      then have "r1 \<cdot> \<tau> = r1 \<cdot> \<sigma>1" by (induct r1) (auto simp: \<tau>_def)
      then show ?thesis using t by auto
    next
      case True
      obtain q' where q':"q' \<in> poss r1" and q'x:"r1 |_ q' = Var x"
        using  supteq_imp_subt_at[OF supteq_Var[OF True]] by auto
      have qr1\<sigma>1:"q' \<in> poss (r1 \<cdot> \<sigma>1)" using q' by simp
      from linear_term_replace_in_subst[OF linr1 q' q'x, of \<sigma>1 \<tau>]
      have "r1 \<cdot> \<tau> = replace_at (r1 \<cdot> \<sigma>1) q' (\<tau> x)" by (simp add: \<tau>_def)
      also have "... = replace_at (r1 \<cdot> \<sigma>1) q' (replace_at (\<sigma>1 x) q2 (r2 \<cdot> \<sigma>2))" by (simp add: \<tau>_def)
      also have "... = replace_at (r1 \<cdot> \<sigma>1) (q' <#> q2) (r2 \<cdot> \<sigma>2)"
        using q' q'x ctxt_of_pos_term_append[OF qr1\<sigma>1] ctxt_compose_apply_distrib by auto
      finally have r1\<tau>:"r1 \<cdot> \<tau> = replace_at (r1 \<cdot> \<sigma>1) (q' <#> q2) (r2 \<cdot> \<sigma>2)" .
      from xq2 q'x have "(r1 \<cdot> \<sigma>1) |_ (q' <#> q2) = l2 \<cdot> \<sigma>2"
        using q' subt_at_append[OF qr1\<sigma>1, of q2] by auto
      then have "r1 \<cdot> \<sigma>1 = (ctxt_of_pos_term (q' <#> q2) (r1 \<cdot> \<sigma>1))\<langle>l2 \<cdot> \<sigma>2\<rangle>"
        using qr1\<sigma>1 ctxt_supt_id \<sigma>1 p2 poss_append_poss q' q'x subt_at_subst q q1 q12 x by metis
      from rstepI[OF lr2 this] r1\<tau> have "(r1 \<cdot> \<sigma>1, r1 \<cdot> \<tau>) \<in> ?R" by simp
      then show ?thesis using t by blast
    qed
    from tcr ucr show ?thesis by blast
  qed
qed

lemma strongly_closed_linear_imp_strongly_confluent:
  assumes "strongly_closed R"
    and "linear_trs R"
  shows "strongly_confluent (rstep R)"
proof
  let ?R = "rstep R"
  fix s t u
  assume "(s, t) \<in> ?R" and "(s, u) \<in> ?R"
  then obtain l1 r1 p1 \<sigma>1 l2 r2 p2 \<sigma>2 
    where st:"(s, t) \<in> rstep_r_p_s R (l1, r1) p1 \<sigma>1"
    and su:"(s, u) \<in> rstep_r_p_s R (l2, r2) p2 \<sigma>2"
    using rstep_iff_rstep_r_p_s by metis
  consider (parallel) "p1 \<bottom> p2" | (nested1) "p1 \<le> p2" | (nested2) "p2 \<le> p1" 
    using parallel_pos by auto
  then show "\<exists>v. (t, v) \<in> ?R\<^sup>* \<and> (u, v) \<in> ?R\<^sup>="
  proof (cases)
    case parallel
    then show ?thesis using parallel_steps st su by (blast dest: rstep_r_p_s_imp_rstep)
  next
    case nested1
    with strongly_closed_nested_pos[OF st su] assms show ?thesis by auto
  next
    case nested2
    with strongly_closed_nested_pos[OF su st] assms show ?thesis by auto
  qed
qed

corollary strongly_closed_linear_CR:
  assumes "strongly_closed R"
    and "linear_trs R"
  shows "CR (rstep R)"
using assms strongly_closed_linear_imp_strongly_confluent strong_confluence_imp_CR by auto

end
