(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2011-2015)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2011-2015)
License: LGPL (see file COPYING.LESSER)
*)
theory List_Order
imports
  "$AFP/Abstract-Rewriting/Abstract_Rewriting"
  "$AFP/Abstract-Rewriting/Relative_Rewriting"
  "../Auxiliaries/Util"
begin

locale pre_order_pair = 
  fixes S :: "'a rel"
    and NS :: "'a rel"
  assumes refl_NS: "refl NS"
    and trans_S: "trans S"
    and trans_NS: "trans NS"
begin

lemma refl_NS_point: "(s,s) \<in> NS" using refl_NS unfolding refl_on_def by blast

lemma rtrancl_NS: "NS^* = NS" 
  by (rule trans_refl_imp_rtrancl_id[OF trans_NS refl_NS])

lemma trans_S_point: "\<And> x y z. (x,y) \<in> S \<Longrightarrow> (y,z) \<in> S \<Longrightarrow> (x,z) \<in> S"
  using trans_S unfolding trans_def by blast

lemma trans_NS_point: "\<And> x y z. (x,y) \<in> NS \<Longrightarrow> (y,z) \<in> NS \<Longrightarrow> (x,z) \<in> NS"
  using trans_NS unfolding trans_def by blast
end

locale compat_pair = 
  fixes S NS :: "'a rel" 
  assumes compat_NS_S: "NS O S \<subseteq> S"
    and compat_S_NS: "S O NS \<subseteq> S"
begin
lemma compat_NS_S_point: "\<And> x y z. (x,y) \<in> NS \<Longrightarrow> (y,z) \<in> S \<Longrightarrow> (x,z) \<in> S"
  using compat_NS_S by blast

lemma compat_S_NS_point: "\<And> x y z. (x,y) \<in> S \<Longrightarrow> (y,z) \<in> NS \<Longrightarrow> (x,z) \<in> S"
  using compat_S_NS by blast

lemma relto_S: "relto S NS = S"
proof -
  let ?R = "relto S NS"
  have one: "S \<subseteq> ?R" by auto
  have two: "?R \<subseteq> S" 
  proof
    fix a d
    assume "(a,d) \<in> ?R"
    then obtain b c where ab: "(a,b) \<in> NS^*" and bc: "(b,c) \<in> S"
      and c_d: "(c,d) \<in> NS^*" by auto
    from ab have ac: "(a,c) \<in> S" 
    proof (induct rule: converse_rtrancl_induct)
      case base
      show ?case by (rule bc)
    next
      case (step a d)
      from compat_NS_S_point[OF step(1) step(3)] show ?case .
    qed
    from c_d show ad: "(a,d) \<in> S"
    proof (induct rule: rtrancl_induct)
      case base
      show ?case by (rule ac)
    next
      case (step b d)
      from compat_S_NS_point[OF step(3) step(2)] show ?case .
    qed
  qed
  from one two show ?thesis by auto
qed  

end

locale order_pair = pre_order_pair S NS + compat_pair S NS 
  for S NS :: "'a rel"

locale SN_ars =
  fixes S :: "'a rel"
  assumes SN: "SN S"

locale SN_pair = compat_pair S NS + SN_ars S for S NS :: "'a rel"

locale SN_order_pair = order_pair S NS + SN_ars S for S NS :: "'a rel"

sublocale SN_order_pair \<subseteq> SN_pair ..

type_synonym 'a list_ext = "'a rel \<Rightarrow> 'a rel \<Rightarrow> 'a list rel"

locale list_order_extension =
  fixes s_list :: "'a list_ext"
   and ns_list :: "'a list_ext"
  assumes extension: "SN_order_pair S NS \<Longrightarrow> SN_order_pair (s_list S NS) (ns_list S NS)"
     and s_map: "\<lbrakk>\<And> a b. (a,b) \<in> S \<Longrightarrow> (f a,f b) \<in> S; \<And> a b. (a,b) \<in> NS \<Longrightarrow> (f a,f b) \<in> NS\<rbrakk> \<Longrightarrow> (as,bs) \<in> s_list S NS \<Longrightarrow> (map f as, map f bs) \<in> s_list S NS" 
     and ns_map: "\<lbrakk>\<And> a b. (a,b) \<in> S \<Longrightarrow> (f a,f b) \<in> S; \<And> a b. (a,b) \<in> NS \<Longrightarrow> (f a,f b) \<in> NS\<rbrakk> \<Longrightarrow> (as,bs) \<in> ns_list S NS \<Longrightarrow> (map f as, map f bs) \<in> ns_list S NS" 
     and all_ns_imp_ns: "length as = length bs \<Longrightarrow> \<lbrakk>\<And> i. i < length bs \<Longrightarrow>  (as ! i, bs ! i) \<in> NS\<rbrakk> \<Longrightarrow> (as,bs) \<in> ns_list S NS"

end
