(*
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2016)
License: LGPL (see file COPYING.LESSER)
*)

section \<open>AC Subterm Criterion -- Implementation\<close>

theory AC_Subterm_Criterion_Impl
imports
  "../Orderings/Subterm_Multiset"
  "../Orderings/Status_Impl"
  "../Auxiliaries/Multiset_Extension_Impl"
  "../Term_Order_Impl"
  "../Framework/AC_Dependency_Pair_Problem_Spec"
  "../AC_Rewriting/AC_Rewriting_Impl"
  AC_Subterm_Criterion
begin

context
  fixes proj :: "'f status"
    and F :: "'f sig"
begin
fun proj_term_impl
where
  "proj_term_impl (Var x) = [Var x]"
| "proj_term_impl (Fun f ts) =
    (if (f, length ts) \<in> F then concat (map (\<lambda>i. proj_term_impl (ts ! i)) (status proj (f, length ts)))
    else [Fun f ts])"

lemma proj_term_impl[simp]: "mset (proj_term_impl t) = proj_term proj F t"
proof (induct t)
  case (Fun f ts)
  show ?case
  proof (cases "(f,length ts) \<in> F")
    case True
    def idx \<equiv> "status proj (f, length ts)"
    def xs \<equiv> "map (\<lambda>i. (proj_term_impl (ts ! i))) idx"
    from True have id: "?thesis = (mset (concat xs) = 
      (\<Union># mset (map (\<lambda>i. proj_term proj F (ts ! i)) idx)))" (is "_ = (_ = ?r)") 
      by (simp add: idx_def xs_def)
    {
      fix i
      assume "i \<in> set idx"
      with status[of proj f "length ts"] have "i < length ts" unfolding idx_def by auto
      hence "ts ! i \<in> set ts" unfolding set_conv_nth by auto
      from Fun[OF this] have "proj_term proj F (ts ! i) = mset (proj_term_impl (ts ! i))" by simp
    } note * = this
    have id2: "map (\<lambda>i. proj_term proj F (ts ! i)) idx = map mset xs" unfolding xs_def
      by (rule nth_equalityI, insert *, auto)
    show ?thesis unfolding id id2 mset_concat_union by simp
  qed auto
qed auto

definition "strict_supt_mul s t = 
  fst (mul_ext (\<lambda> x y. (x \<rhd> y, x = y)) (proj_term_impl s) (proj_term_impl t))"

definition "weak_supt_mul s t = 
  snd (mul_ext (\<lambda> x y. (x \<rhd> y, x = y)) (proj_term_impl s) (proj_term_impl t))"

private lemma Id_unfold: "Id = {(x, y). x = y}" by auto

lemma strict_supt_mul[simp]: "strict_supt_mul s t = suptproj_pred proj F s t"
  unfolding strict_supt_mul_def mul_ext_def Let_def by (auto simp: Id_unfold)

lemma weak_supt_mul[simp]: "weak_supt_mul s t = suptprojeq_pred proj F s t"
  unfolding weak_supt_mul_def mul_ext_def Let_def by (auto simp: Id_unfold)
end  


definition check_suptproj_pred :: "'f status \<Rightarrow> 'f sig \<Rightarrow> ('f :: show, 'v :: show)rule \<Rightarrow> shows check" where
  "check_suptproj_pred \<pi> F lr = check (case lr of (l,r) \<Rightarrow> strict_supt_mul \<pi> F l r) 
     (shows ''could not orient rule '' o shows_rule lr o shows '' by supt^mul'')"

definition check_supteqproj_pred :: "'f status \<Rightarrow> 'f sig \<Rightarrow> ('f :: show, 'v :: show)rule \<Rightarrow> shows check" where
  "check_supteqproj_pred \<pi> F lr = check (case lr of (l,r) \<Rightarrow> weak_supt_mul \<pi> F l r) 
     (shows ''could not orient rule '' o shows_rule lr o shows '' by supteq^mul'')"

lemma check_suptproj_pred[simp]: "isOK(check_suptproj_pred \<pi> F lr) = (suptproj_pred \<pi> F (fst lr) (snd lr))"
  unfolding check_suptproj_pred_def by auto

lemma check_supteqproj_pred[simp]: "isOK(check_supteqproj_pred \<pi> F lr) = (suptprojeq_pred \<pi> F (fst lr) (snd lr))"
  unfolding check_supteqproj_pred_def by auto

definition
  ac_subterm_proc ::
    "('dpp, 'f :: show, 'v :: show) ac_dpp_ops \<Rightarrow> 'f status_impl \<Rightarrow>
     ('f,'v)rules \<Rightarrow> 'dpp proc"
where
  "ac_subterm_proc I pi P_remove dpp = check_return (do {
     let P = ac_dpp_ops.pairs I dpp;
     let R = ac_dpp_ops.rules I dpp;
     let E = ac_dpp_ops.E I dpp;
     let RE = R @ E;
     let F = map fst pi;
     let FF = set F;
     let pi_opt = status_of pi;
     check (pi_opt \<noteq> None) (shows ''argument filter lists invalid positions'');
     let \<pi> = the pi_opt;
     let Premove = set P_remove;
     let (ps, pns) = partition (\<lambda> lr. lr \<in> Premove) P;
     check_allm (\<lambda> f. check (status \<pi> f \<noteq> []) 
       (shows ''status of symbol '' o shows f o shows '' in F must be non-empty'')) F;
     check_size_preserving_trs E 
       <+? (\<lambda> s. shows ''E is not size preserving\<newline>'' o s);
     check_allm (\<lambda> (l,r). check (is_Fun l) (shows ''variables as lhss not allowed'')) RE;        
     check_allm (check_supteqproj_pred \<pi> FF) (filter (\<lambda> lr. the (root (fst lr)) \<in> FF) RE)
       <+? (\<lambda>s. shows ''problem when orienting rules with root in F\<newline>'' o s);
     check_allm (check_supteqproj_pred \<pi> FF) pns
       <+? (\<lambda>s. shows ''problem when orienting DPs\<newline>'' o s);
     check_allm (check_suptproj_pred \<pi> FF) ps
       <+? (\<lambda>s. shows ''problem when orienting DPs\<newline>'' o s)
   } <+? (\<lambda>s. shows ''could not apply the AC subterm processor\<newline>'' o s))
   (ac_dpp_ops.delete_pairs_rules I dpp P_remove [])"

lemma (in ac_dpp_spec) ac_subterm_proc:
  shows "sound_proc_impl (ac_subterm_proc I pi ps)"
proof 
  fix d d'
  assume fin: "finite_ac_dpp (ac_dpp d')"
    and ok: "ac_subterm_proc I pi ps d = return d'"
  def F \<equiv> "map fst pi"
  def \<pi> \<equiv> "the (status_of pi)"
  let ?P = "set (P d)"
  let ?Pw = "set (Pw d)"
  let ?PP = "?P \<union> ?Pw"
  let ?R = "set (R d)"
  let ?Rw = "set (Rw d)"
  let ?E = "set (E d)"
  let ?RE = "?R \<union> ?Rw \<union> ?E"
  let ?Sup = "suptproj_pred \<pi> (set F)"
  let ?SupEq = "suptprojeq_pred \<pi> (set F)"
  obtain Ps Pns where p: "partition (\<lambda> lr. lr \<in> set ps) (pairs d) = (Ps,Pns)" by force
  note ok = ok[unfolded ac_subterm_proc_def Let_def p split, folded F_def \<pi>_def, simplified]
  from ok have var: "\<And> l r. (l, r)\<in> ?RE \<Longrightarrow> is_Fun l"
      and size: "size_preserving_trs (set (E d))"
      and d': "d' = ac_dpp_ops.delete_pairs_rules I d ps []"
      and \<pi>: "\<forall>f\<in> set F. status \<pi> f \<noteq> []"
      and supt: "\<And> l r. (l,r) \<in> set Ps \<Longrightarrow> ?Sup l r"
      and supteq: "\<And> l r. (l,r) \<in> set Pns \<Longrightarrow> ?SupEq l r"
    by auto
  {
    fix l r
    assume "(l,r) \<in> ?PP"
    with p have "(l,r) \<in> set Ps \<union> set Pns" by auto
    with s_ns_mul_ext[OF supt[of l r]] supteq[of l r] have "?SupEq l r" by auto
  } note PP = this
  {
    fix l r
    assume lr: "(l,r) \<in> ?RE" and "root l \<in> Some ` set F"
    with ok have rt: "the (root l) \<in> set F" by (cases l, auto)
    from ok have "(l,r) \<in> ?RE \<Longrightarrow> the (root (fst (l,r))) \<in> set F \<Longrightarrow> ?SupEq (fst (l,r)) (snd (l,r))"
      by blast
    with rt lr have "?SupEq l r" by auto
  } note RE = this
  have id: "?PP = set (pairs d)" "?RE = set (R d @ Rw d @ E d)" by auto
  note fin = fin[unfolded d' delete_pairs_rules_sound, simplified]
  show "finite_ac_dpp (ac_dpp d)" unfolding ac_dpp_sound
    by (rule ac_subterm_proc[OF \<pi> fin PP RE supt size var], insert p, auto)
qed
end